/*
 * (c) Unicontrollers, Ltd., 2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file qc.c
 * The main file of the command line tool named qc, which
 * is used for Quality Control tests
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>

#include "uncusb.h"
#include "unc_data.h"

#define FACTORY_ID 0x7775

#define SLOT_IN_ID 0x04
#define SLOT_OUT_ID 0x0f

static void press_enter()
{
    int c;
    printf("# press enter (or Ctrl-C to abort)");
    while((c = getchar()) != EOF && c != '\n') {}
}

static void detect_hardware(void *dev_hdl, int *slot1, int *slot2)
{
    struct unc001_global_configuration_str conf;
    int r = unc001_get_configuration(dev_hdl, &conf);
    if(r == -1) {
        fprintf(stderr, "get_configuration failed\n");
        printf("@ FAILED\n");
        exit(1);
    }
    *slot1 = conf.hardware_config & 0x0f;
    *slot2 = (conf.hardware_config & 0xf0) >> 4;
}

static const char *slot_status_msg(int code)
{
    switch(code) {
        case 0x00:         return "EMPTY  ";
        case SLOT_IN_ID:   return "INPUT  ";
        case SLOT_OUT_ID:  return "OUTPUT ";
        default:           return "UNKNOWN";
    }
}

static const char * const test_action_seq[] = {
    "----",
    "++++",
    "----",
    "++++",
    "----",
    "++++",
    "----",
    "---+",
    "--+-",
    "-+--",
    "+---",
    "----",
    "+---",
    "-+--",
    "--+-",
    "---+",
    "----",
    "+---",
    "----",
    "-+--",
    "----",
    "--+-",
    "----",
    "---+",
    "----",
    "+++-",
    "++-+",
    "+-++",
    "-+++",
    "++++",
    "-+++",
    "+-++",
    "++-+",
    "+++-",
    "----",
    "+-+-",
    "-+-+",
    "+-+-",
    "-+-+",
    "+-+-",
    "-+-+",
    "++--",
    "--++",
    "++--",
    "--++",
    "++--",
    "--++",
    "+--+",
    "-++-",
    "+--+",
    "-++-",
    "+--+",
    "-++-",
    "----",
    NULL
};

static const int test_action_count =
    sizeof(test_action_seq) / sizeof(*test_action_seq);

static void parse_action(const char *s, int *action, int *bits)
{
    int i;
    *action = 0;
    *bits = 0;
    for(i = 0; i < 4; i++) {
        *action <<= 2;
        *bits <<= 1;
        if(s[i] == '+') {
            *action |= 1;
            *bits |= 1;
        } else {
            *action |= 2;
        }
    }
}

static void make_stat_str(int bits, char s[16])
{
    int i;
    for(i = 0; i < 4; i++) {
        s[i*3]   = '[';
        s[i*3+1] = (bits & (1 << i)) ? '*' : ' ';
        s[i*3+2] = ']';
    }
    s[12] = 0;
}

static int get_dev_stat(void *dev_hdl)
{
    struct unc001_global_state_str state;
    int r = unc001_get_fullstate(dev_hdl, &state);
    if(r == -1) {
        fprintf(stderr, "get_global_state failed\n");
        printf("@ FAILED\n");
        exit(1);
    }
    return state.state_bits;
}

static int manual_output_test = 1;

static void test_output(void *dev_hdl, int group)
{
    const char * const * p;
    int tn = 1;
    printf("# testing output group %d\n", group);
    for(p = test_action_seq; *p; p++) {
        int action, bits, sb;
        char stat_exp[16], stat_real[16];
        parse_action(*p, &action, &bits);
        unc001_perform_action(dev_hdl, (group << 8) | action);
        usleep(10000); /* XXXXXXXXXXXXX */
        sb = get_dev_stat(dev_hdl);
        if(group > 0)
            sb >>= ((group+1) * 4);
        sb &= 0x0f;
        make_stat_str(bits, stat_exp);
        make_stat_str(sb, stat_real);
        printf("# Test %02d/%02d: %s | %s\n",
            tn++, test_action_count, stat_exp, stat_real);
        if(sb != bits) {
            printf("# Device returned unexpected lines status\n");
            printf("@ FAILED OUTSTAT\n");
        }
        if(manual_output_test)
            press_enter();
        else
            usleep(500000);
    }
}

static void test_input(void *dev_hdl, int slot)
{
    int imin = (slot == 1) ? 0 : 4;
    int imax = (slot == 1) ? 3 : 7;
    int i;
    int counters[8];
    printf("# testing input group %d\n", slot);
    for(i = imin; i <= imax; i++) {
        unc001_set_reactions(dev_hdl, i, 0x70000000);
    }
    unc001_zero_query_counters(dev_hdl, counters, 8);
    printf("Please press CH1, CH2, CH3, CH4 one time each\n");
    press_enter();
    unc001_zero_query_counters(dev_hdl, counters, 8);
    for(i = imin; i <= imax; i++) {
        if(counters[i] == 2) {
            printf("@ SUCCESS INP %d\n", i);
        } else {
            printf("@ FAILED INP %d : %d instead of 2\n", i, counters[i]);
        }
    }
}

static void test_slot(void *dev_hdl, int slot, int mod_id)
{
    switch(mod_id) {
        case 0x00:
            break;
        case SLOT_IN_ID:
            test_input(dev_hdl, slot);
            break;
        case SLOT_OUT_ID:
            test_output(dev_hdl, slot);
            break;
        default:
            printf("# unknown module in slot %d, please doublecheck!\n", slot);
            printf("@ FAILED UNKMOD %d:%d\n", slot, mod_id);
    }
}

static int skip_main_group = 0;
static int skip_module_1 = 0;
static int skip_module_2 = 0;

static void help()
{
    printf("Usage:    qc [options]\n");
    printf("      -b     perform output tests with no manual intervention\n");
    printf("      -0     skip main group test\n");
    printf("      -1     skip module 1 test\n");
    printf("      -2     skip module 2 test\n");
    printf("      -h     print this help and exit\n");
}

static void parse_cmdline(int argc, char **argv)
{
    int c;
    while(-1 != (c = getopt(argc, argv, "b012h"))) {
        switch(c) {
            case 'b':
                manual_output_test = 0;
                break;
            case '0':
                skip_main_group = 1;
                break;
            case '1':
                skip_module_1 = 1;
                break;
            case '2':
                skip_module_2 = 1;
                break;
            case 'h':
            default:
                help();
        }
    }
}

int main(int argc, char **argv)
{
    struct unc001_item *list;
    void *dh = NULL;
    int ser1, ser2;
    int slot1, slot2;

    parse_cmdline(argc, argv);

    if(geteuid()!=0) {
        fprintf(stderr,
               "WARNING: usually only root can operate this program\n");
    } 

    unc001_usb_init();
    list = unc001_scan(FACTORY_ID);
    if(list) {
        dh = list->dev_hdl;
        ser1 = list->sernum[0];
        ser2 = list->sernum[1];
        list->dev_hdl = NULL;
        unc001_destroy_list(list);
    } else {
        fprintf(stderr, "No devices found, ");
        fprintf(stderr, "make sure it is connected and has the id %x\n",
                        FACTORY_ID);
        return 1;
    }
    if(!dh) {
        fprintf(stderr, "The device is incompatible (very strange)\n");
        return 1;
    }
    printf("# Device is found\n");
    printf("@ SERIAL: %x:%x\n", ser1, ser2);
    detect_hardware(dh, &slot1, &slot2);
    printf("@ SLOT1: %s\n", slot_status_msg(slot1));
    printf("@ SLOT2: %s\n", slot_status_msg(slot2));
    printf("# Please check the above info for correctness\n");
    press_enter();
    if(!skip_main_group) test_output(dh, 0);
    if(!skip_module_1) test_slot(dh, 1, slot1);
    if(!skip_module_2) test_slot(dh, 2, slot2);
    return 0;
}
