/*
 * UNC001 host-side software. (c) Unicontrollers, Ltd., 2010-2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file uncctl.c
 * The main file of the command line tool named uncctl, which
 * allows to control the UNC001 device in batch mode
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdio.h>
#ifdef WIN32
#include <getopt.h>
#include <conio.h>
#else
#include <unistd.h>
#endif
#include <string.h>
#include <stdlib.h>

#include <stdint.h>


#include "uncusb.h"
#include "unc_data.h"

static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static void help()
{
    static const char *htext[] = {
        "uncctl vers. 2.2    (c) Unicontrollers Ltd., 2010-2012",
        "usage: uncctl <opmode> <options>",
        "  opmodes: -L            find and list UNC0xx devices",
        "           -Q            dump the full state & configuration",
        "           -Z            query and zero input counters",
        "           -D            dump eeprom (for debug only)",
        "           -T <new id>   set UNC ID for the device",
        "           -G <conf_reg> change configuration register",
        "           -H <hw_conf> \"really\"   change HW CONF (dangerous!)",
        "           -A <action>   perform \"action\"",
        "           -R <in> <val> set reaction register for the <in> (0..7)",
        "                      to <val> (hex number or 'num:action:action')",
        "           -W            write configuration to eeprom",
        "           -E            restore configuration from eeprom",
        "           -F            restore factory defaults",
        "           -Y <nn> <hex_data> set Enn register's contents", 
        "                      <hex_data> must be exactly 16 hex-digits",
        "           -M            send a text message to the device's COM",
        "           -C            intercept the device's COM completely,",
        "                      start a kind of chat with it",
#ifdef WIN32
        "                      NOT RELIABLE under Win32 systems",
#endif
        "           -I            display the input stream from the device",
        "           -K            clear message buffer in the glob. state",
        "",
        "  options: -i <id>       use the device with this id",
        "                 (otherwise, the first found one is used)",
        "           -t <msec>     set timeout (ms), default 5000",
        "           -n            with -M, don't send the newline char",
        "           -v            be verbose (useful with -Q)",
        "",
        "NOTE: Device ids must be given as hex word, e.g. 21af.",
        "      Action may be specified as hex (with leading `x'),",
        "      binary or four action chars (+,-,^,=), where 01 or +",
        "      means turn on, 10 or - means turn off, 11 or ^ means",
        "      switch to the opposit state and 00 or = means no action,",
        "      so \"x63\", \"01100011\" and \"+-=^\" are the same:",
        "      lineD on, lineC off, leave lineB as is, switch lineA",
        "   some examples of actions:",
        "      xAA    10101010    ----     all lines off",
        "      x55    01010101    ++++     all lines on",
        "      xff    11111111    ^^^^     switch all lines",
        "      x00    00000000    ====     no operation",
        "      x04    00000100    ==+=     lineB on, don't touch others",
        "  You can use actions with -R as well, e.g., \"-R 3 6000:++++:----\"",
        "  is the same as \"-R 3 600055AA\"",
        NULL
    };
    int i;
    for(i=0; htext[i]; i++)
        printf("%s\n", htext[i]);
}

/* command line parameters */

static enum {
    opm_undef,
    opm_list,
    opm_query,
    opm_zero,
    opm_dump_eeprom,
    opm_setid,
    opm_set_conf,
    opm_set_hwconf,
    opm_write_eeprom,
    opm_restore_from_eeprom,
    opm_restore_factory,
    opm_write_e_reg,
    opm_clear_messagebuf,
    opm_input_stream,
    opm_action,
    opm_message,
    opm_intercept_com,
    opm_set_reactions
} operation_mode = opm_undef;
int device_id = 0;
int be_verbose = 0;
char **rest_args;
char *action_arg;
int send_newline = 1;


static int set_op_mode(int m)
{
    if(operation_mode != opm_undef) {
        fprintf(stderr, "You can only specify ONE operation mode\n");
        return 0;
    }
    operation_mode = m;
    return 1;
}

static int parse_cmd_line(int argc, char **argv)
{
    int c;
    while(-1 != (c = getopt(argc, argv, "LQZDTGHARWEFYCMKIi:t:nv"))) {
        switch(c) {
            case 'L':
                if(!set_op_mode(opm_list)) return 0;
                break;
            case 'Q':
                if(!set_op_mode(opm_query)) return 0;
                break;
            case 'Z':
                if(!set_op_mode(opm_zero)) return 0;
                break;
            case 'D':
                if(!set_op_mode(opm_dump_eeprom)) return 0;
                break;
            case 'T':
                if(!set_op_mode(opm_setid)) return 0;
                break;
            case 'G':
                if(!set_op_mode(opm_set_conf)) return 0;
                break;
            case 'H':
                if(!set_op_mode(opm_set_hwconf)) return 0;
                break;
            case 'A':
                if(!set_op_mode(opm_action)) return 0;
                action_arg = argv[optind];
                optind++;
                break;
            case 'R':
                if(!set_op_mode(opm_set_reactions)) return 0;
                break;
            case 'W':
                if(!set_op_mode(opm_write_eeprom)) return 0;
                break;
            case 'E':
                if(!set_op_mode(opm_restore_from_eeprom)) return 0;
                break;
            case 'F':
                if(!set_op_mode(opm_restore_factory)) return 0;
                break;
            case 'Y':
                if(!set_op_mode(opm_write_e_reg)) return 0;
                break;
            case 'M':
                if(!set_op_mode(opm_message)) return 0;
                break;
            case 'C':
                if(!set_op_mode(opm_intercept_com)) return 0;
                break;
            case 'K':
                if(!set_op_mode(opm_clear_messagebuf)) return 0;
                break;
            case 'I':
                if(!set_op_mode(opm_input_stream)) return 0;
                break;
            case 'i': {
                char *err;
                int i = strtol(optarg, &err, 16);
                if(*optarg != '\0' && *err == '\0') {
                    device_id = i;
                } else {
                    fprintf(stderr, "Invalid device id\n");
                    return 0;
                }
                break;
            }
            case 't': {
                char *err;
                int i = strtol(optarg, &err, 10);
                if(*optarg != '\0' && *err == '\0') {
                    unc001_usb_timeout = i;
                } else {
                    fprintf(stderr, "Invalid timeout value\n");
                    return 0;
                }
                break;
            }
            case 'n':
                send_newline = 0;
                break;
            case 'v':
                be_verbose++;
                break;
            default:
                return 0;
        }
    }
    rest_args = argv + optind;
    if(operation_mode == opm_undef) {
        fprintf(stderr, "You must specify the operation mode\n");
        return 0;
    }
    return 1;
}

static void print_devices(struct unc001_item *list)
{
    int n = 0;
    if(!list) {
        printf("No UNC01 devices found\n");
        return;
    }
    printf("<id>  [serial number ]   version\n");
    for(; list; list=list->next) {
        printf("%04x  [%04x:%04x:%04x]  %8d %s\n",
               list->sernum[2],
               list->sernum[0],
               list->sernum[1],
               list->sernum[2],
               list->version,
               (list->version==0 || !list->dev_hdl) ? "DEPRECATED":"");
        n++;
    }
    printf("%d UNC0xx device(s) found\n", n);
}

static void verbose_action_print(const char *whatfor, int a)
{
    int i;
    static const char ach[4] = { '=', '+', '-', '^' };
    static const char *act[4] = { "keep  ", "on    ", "off   ", "toggle" };
    printf("   action on %s ", whatfor);
    for(i=3; i>=0; i--)
        printf("%c", ach[a>>(i*2) & 3]);
    printf("    ");
    for(i=3; i>=0; i--)
        printf("%d:%s ", i, act[a>>(i*2) & 3]);
    printf("\n");
}

static int query_device(void *dev_hdl)
{
    struct unc001_global_configuration_str conf;
    struct unc001_global_state_str stat;
    struct unc001_extra_state_str extra;
    int first_input, last_input;
    int i, r;

    r = unc001_get_configuration(dev_hdl, &conf);
    if(r == -1) {
        fprintf(stderr, "get_configuration failed\n");
        //return 1;
    }
    r = unc001_get_fullstate(dev_hdl, &stat);
    if(r == -1) {
        fprintf(stderr, "get_fullstate failed\n");
        return 1;
    }

    first_input = ((conf.hardware_config & 0x0f) == 0x04) ? 0 : 4;
    last_input  = ((conf.hardware_config & 0xf0) == 0x40) ? 7 : 3;

    printf("version: %d\n"
           "hardware configuration: %08x user configuration: %08x\n",
           conf.version, conf.hardware_config, conf.user_config);
    printf("state register: %04x", stat.state_bits);
    if(conf.version >= 2120203)
        printf("  timemark: %09d\n", stat.timemark);
    else
        printf("\n");
    if(be_verbose) {
        printf("   ");
        for(i=last_input; i>=first_input; i--)
            printf("in%d:%s ", i, ((stat.state_bits>>(i+8))&1) ? "on ":"off");
        for(i=3; i>=0; i--)
            printf("out%d:%s ", i, ((stat.state_bits>>i)&1) ? "on ":"off");
        printf("\n");
    }
    for(i = first_input; i<=last_input; i++)
    {
        if(conf.reactions[i] & 0x80000000) /* analog mode */
            continue;
        printf("in%d: mode %08x  counter %d\n",
                i, conf.reactions[i], stat.counters[i]);
        if(be_verbose) {
            static const char *icm[] =
                { "none", "press", "release", "press&release" };
            printf("   increment counter on: %s\n",
                   icm[(conf.reactions[i]>>29)&3]);
            verbose_action_print("release:", (conf.reactions[i]>>8)&0xff);
            verbose_action_print("press:  ", (conf.reactions[i])   &0xff);
        }
    }
    if(stat.buf_start != stat.buf_end) {
        printf("message: <[");
        stat.buf_end &= 0x1f;
        stat.buf_start &= 0x1f;
        while(stat.buf_start != stat.buf_end) {
            putchar(stat.message_buffer[stat.buf_start]);
            ++stat.buf_start;
            stat.buf_start &= 0x1f;
        }
        printf("]>\n");
    } 
    if(conf.version >= 2120203) {
        r = unc001_get_extrastate(dev_hdl, &extra);
        if(r == -1) {
            fprintf(stderr, "get_extrastate failed\n");
            return 1;
        }
        if(extra.data_buffer[0] & 0x07) {
            for(i = 1; i < (extra.data_buffer[0] & 0x07); i++)
                printf("%02x ", extra.data_buffer[i]);
            printf("\n");
        }
        for(i=0; i<4; i++) {
            if(extra.ow_ids[i][0]) {
                int j, temp, grad, migrad;
                char tsign = '+';
                printf("1w: ");
                for(j=7; j>=0; j--) printf("%02x", extra.ow_ids[i][j]);
                temp = extra.ow_temperature[i];
                switch(temp) {
                case (short)0x8000: /* no temperature here */
                    printf("\n");
                    break;
                case 0x7fff: /* crc error */
                    printf("    temperature: CRC error\n");
                    break;
                default:
                    grad = temp >> 4;
                    migrad = (temp & 0x0f) * 625;
                    if(temp < 0) {
                        tsign = '-';
                        if(migrad > 0) {
                            migrad = 10000 - migrad;
                            grad = -grad - 1;
                        } else {
                            grad = -grad;
                        }
                    }
                    printf("    temp: (0x%04hx) %c%d.%04d C\n",
                                     temp, tsign, grad, migrad);
                }
            }
        }
    }


    return 0;
}

static int zero_query_counters(void *dev_hdl)
{
    int cnt[UNC_INPUTS_COUNT];
    int r, i;
    r = unc001_zero_query_counters(dev_hdl, cnt, UNC_INPUTS_COUNT);
    if(r < 0) {
        fprintf(stderr, "error: %s\n", unc001_strerror());
        return 1;
    }
    for(i=0; i<r; i++) {
        printf("%d ", cnt[i]);
    }
    printf("\n");
    return 0;
}

static void hexdump(char *buffer, int len, int base)
{
    int i;
    for(i = 0; i < len; i++){
        if(i % 16 == 0){
            if(i != 0) printf("\n");
            printf("%03x:  ", base + i);
        } else
        if(i % 8 == 0) {
            printf("   ");
        } else
        if(i % 4 == 0) {
            printf("  ");
        } else {
            printf(" ");
        }
        printf("%02x", buffer[i] & 0xff);
    }
    if(i != 0)
        printf("\n");
}

static int dump_eeprom(void *dev_hdl, int version)
{
    char buf[128];
    int r, k;
    for(k = 0; k < ((version >= 2120203) ? 4 : 1); k++) {
        r = unc001_dump_eeprom(dev_hdl, buf);
        if(r <= 0) return 1;
        hexdump(buf, sizeof(buf), k*sizeof(buf));
    }
    return 0;
}

static int generic_set_command(void *dev_hdl, char opt, int check,
                               int (*fun)(void*, unsigned int))
{
    char *err;
    int r;
    long id;
    if(check) {
        int ok = 1;
        if(!rest_args || !rest_args[0] || !rest_args[1])
            ok = 0;
        else
            ok = (strcmp(rest_args[1], "really") == 0);
        if(!ok) {
            fprintf(stderr, "Don't use -%c. "
                "The device can be physically damaged with this option.\n"
                "If you know what you do, specify the word 'really'\n",
                opt);
            return 1;
        }
    }
    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No number specified for -%c\n", opt);
        return 1;
    }
    id = strtol(*rest_args, &err, 16);
    if(*rest_args != '\0' && *err == '\0') {
        /* okay */
        r = fun(dev_hdl, id);
        return r==0 ? 0 : 1;
    } else {
        fprintf(stderr, "Invalid parameter (must be a hex number)\n");
        return 1;
    }
}

static int set_new_id(void *dev_hdl)
{
    return generic_set_command(dev_hdl, 'T', 0, unc001_set_new_id);
}

static int set_config(void *dev_hdl)
{
    return generic_set_command(dev_hdl, 'G', 0, unc001_set_configuration);
}

static int set_hwconfig(void *dev_hdl)
{
    return generic_set_command(dev_hdl, 'H', 1, unc001_set_hardware_conf);
}

static int extract_action_bin(char *databuf, uint16_t *res)
{
    int action = 0;
    while(*databuf) {
        action *= 2;
        switch(*databuf) {
        case '1':
            action += 1;
            /* intentionally no break here */
        case '0':
            break;
        default:
            return 0;
        }
        databuf++;
    }
    *res = action & 0xff;
    return 1;
}

static int extract_action_hex(char *databuf, uint16_t *res)
{
#if 0
    int d1, d2;
    d1 = hexdig2num(databuf[0]);
    if(d1 != -1) {
        d2 = hexdig2num(databuf[1]);
        if(d2 != -1 && databuf[2] == '\0') {
            *res = d1<<4 | d2;
            return 1;
        }
    }
    return 0;
#else
    char *err;
    *res = strtol(databuf, &err, 16);
    return *res && !*err;
#endif
}

static int extract_action_mnem(char *databuf, uint16_t *res)
{
    int action = 0;
    while(*databuf) {
        action <<= 2;
        switch(*databuf) {
        case '=':
            break;       /* 00 == keep untouched */
        case '^':
            action |= 3; /* 11 == switch state */
            break;
        case '+':
            action |= 1; /* 01 == set on */
            break;
        case '.':
        case '-':
            action |= 2; /* 10 == set off */
            break;
        default:
            return 0;
        }
        databuf++;
    }
    *res = action & 0xff;
    return 1;
}

static int extract_action(char *databuf, uint16_t *res)
{
    switch(databuf[0]) {
        case '0':
        case '1':
            return extract_action_bin(databuf, res);
        case 'x':
        case 'X':
            return extract_action_hex(databuf+1, res);
        case '.':
        case '+':
        case '-':
        case '^':
        case '=':
            return extract_action_mnem(databuf, res);
        default:
            return 0;
    }
}


static int perform_action(void *dev_hdl)
{
    uint16_t action = 0; /* 0 is to make stupid compiler happy */
    uint16_t agrp;
    int r;
    char *err;
    if(!action_arg) {
        fprintf(stderr, "No action specified for -A\n");
        return 1;
    }
    if(strchr(action_arg, ':')) {
        char *p = strrchr(action_arg, ':');
        if(!extract_action(p+1, &action)) {
            fprintf(stderr, "Invalid action\n");
            return 1;
        }
        *p = 0;
        agrp = strtoul(action_arg, &err, 16) << 8;
        if(*action_arg == '\0' || *err != '\0') {
            fprintf(stderr, "Invalid action group\n");
            return 1;
        }
        action &= 0x0ff;
        action |= agrp;
    } else {
        if(!extract_action(action_arg, &action)) {
            fprintf(stderr, "Invalid action\n");
            return 1;
        }
    }
    r = unc001_perform_action(dev_hdl, action);
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc001_strerror());
        return 1;
    }
    return 0;
}

static int set_reactions(void *dev_hdl)
{
    char *err;
    int r;
    int line;
    unsigned long reaction;

    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No input line number specified for -R\n");
        return 1;
    }
    if(rest_args[0][1]!='\0' || rest_args[0][0]>'7' || rest_args[0][0]<'0'){
        fprintf(stderr, "Invalid input line (must be 0, 1, ..., 7)\n");
        return 1;
    }
    line = rest_args[0][0] - '0';
    if(!rest_args[1]) {
        fprintf(stderr, "No reaction value specified for -R\n");
        return 1;
    }
    if(strchr(rest_args[1], ':')) {
        char *p = strrchr(rest_args[1], ':');
        uint16_t ac_press, ac_release;
        if(!extract_action(p+1, &ac_press)) {
            fprintf(stderr, "Invalid action string %s\n", p+1);
            return 1;
        }
        *p = 0;
        p = strrchr(rest_args[1], ':');
        if(!extract_action(p ? p+1 : rest_args[1], &ac_release)) {
            fprintf(stderr, "Invalid action string %s\n",
                            p ? p+1 : rest_args[1]);
            return 1;
        }
        if(p) {
            *p = 0;
            reaction = strtoul(rest_args[1], &err, 16) << 16;
            if(*rest_args[1] == '\0' || *err != '\0') {
                fprintf(stderr, "Invalid reaction value\n");
                return 1;
            }
        } else {
            // only one ':' means there's no flags part, should use 0x6000 
            reaction = 0x60000000;
        }
        reaction |= ((unsigned long) ac_press & 0xff);
        reaction |= (((unsigned long) ac_release) & 0xff) << 8;
    } else {
        reaction = strtoul(rest_args[1], &err, 16);
        if(*rest_args[1] == '\0' || *err != '\0') {
            fprintf(stderr, "Invalid reaction value\n");
            return 1;
        }
    }
    if(!(reaction & 0x60000000)) {
        fprintf(stderr, "Warning: counter disabled\n");
    }
    r = unc001_set_reactions(dev_hdl, line, reaction);
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc001_strerror());
        return 1;
    }
    return 0;
}

static int write_e_reg(void *dev_hdl)
{
    char *err, *p;
    int r, i;
    int regnum;
    char data[8];

    if(!rest_args || !*rest_args) {
        fprintf(stderr, "No E reg number specified for -Y\n");
        return 1;
    }
    regnum = strtol(*rest_args, &err, 10);
    if(!**rest_args || *err || regnum < 0 || regnum > 47){
        fprintf(stderr, "Invalid E register number, must be 0..47\n");
        return 1;
    }
    if(!rest_args[1]) {
        fprintf(stderr, "No data specified for -Y\n");
        return 1;
    }
    p = rest_args[1];
    for(i=0; i<8; i++) {
        int d1 = hexdig2num(*p++);
        int d2 = hexdig2num(*p++);
        if(d1 == -1 || d2 == -1) {
            fprintf(stderr, "Invalid data for -Y\n");
            return 1;
        }
        data[7-i] = (d1 << 4) | d2;
    }
    r = unc001_write_e_reg(dev_hdl, regnum, data);
    if(r != 0) {
        fprintf(stderr, "error: %s\n", unc001_strerror());
        return 1;
    }
    return 0;
}

static int input_stream(void *dev_hdl)
{
    int rc;
    char buf[8];
    if(unc001_takeover_stream(dev_hdl) < 0) {
        fprintf(stderr, "Error taking over the stream\n");
        return 1;
    }
    while((rc = unc001_read_stream(dev_hdl, buf)) > 0) {
        int i;
        for(i=0; i<rc; i++)
            putchar(buf[i]);
        fflush(stdout);
    }
    return rc == 0 ? 0 : 1;
}

extern int intercept_com(void *dev_hdl);

int main(int argc, char **argv)
{
    struct unc001_item *list;

    if(argc<2 || !parse_cmd_line(argc, argv)) {
        help();
        return 1;
    }

#ifndef WIN32
    if(geteuid()!=0) {
        fprintf(stderr,
               "WARNING: usually only root can operate this program\n");
    } 
#endif

    unc001_usb_init();

    list = unc001_scan(device_id);
    if(!list) {
        fprintf(stderr, "No devices found\n");
        return 1;
    } else {
        if(operation_mode != opm_list && !list->dev_hdl) {
            fprintf(stderr, "The device is incompatible, check versions\n");
            return 1;
        }
        switch(operation_mode) {
            case opm_undef:
                return 1; /* never happens */
            case opm_list:
                print_devices(list);
                return list ? 0 : 1;
            case opm_query:
                return query_device(list->dev_hdl);
            case opm_zero:
                return zero_query_counters(list->dev_hdl);
            case opm_dump_eeprom:
                return dump_eeprom(list->dev_hdl, list->version);
            case opm_setid:
                return set_new_id(list->dev_hdl);
            case opm_set_conf:
                return set_config(list->dev_hdl);
            case opm_set_hwconf:
                return set_hwconfig(list->dev_hdl);
            case opm_action:
                return perform_action(list->dev_hdl);
            case opm_set_reactions:
                return set_reactions(list->dev_hdl);
            case opm_write_eeprom:
                return unc001_save_to_eeprom(list->dev_hdl)==0?0:1;
            case opm_restore_from_eeprom:
                return unc001_restore_from_eeprom(list->dev_hdl)==0?0:1;
            case opm_restore_factory:
                return unc001_restore_factory(list->dev_hdl)==0?0:1;
            case opm_write_e_reg:
                return write_e_reg(list->dev_hdl);
            case opm_clear_messagebuf:
                return unc001_clear_msg(list->dev_hdl)==0?0:1;
            case opm_input_stream:
                if(list->version < 2120131) {
                    fprintf(stderr, "Device's firmware is too old. ");
                    fprintf(stderr, "It doesn't support this mode.\n");
                    fprintf(stderr, "Consider updating it\n");
                    return 1;
                } else {
                    return input_stream(list->dev_hdl);
                }
            case opm_message:
                if(!rest_args || !*rest_args) {
                    fprintf(stderr, "No message specified for -M\n");
                    return 1;
                }
                {
                    int res;
                    res = unc001_display_message(list->dev_hdl, *rest_args);
                    if(res == -1) return 1;
                    res = send_newline ?
                        unc001_display_message(list->dev_hdl, "\n") : 0;
                    return res == 0 ? 0 : 1;
                }
            case opm_intercept_com:
                return intercept_com(list->dev_hdl);
        }
    }
    return 0;
}
