/*
 * UNC001 host-side software. (c) Unicontrollers, Ltd., 2010-2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file uncusb.c
 * Set of public functions for communicating with the device over USB
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include <stdint.h>

#include <string.h>

#ifdef _WIN32
#define USE_HIDAPI
#endif

#ifndef USE_HIDAPI
#include <usb.h>
#else
#include <stdlib.h>   /* we need malloc/free */
#include <windows.h>
#include <setupapi.h>
#include "hidapi.h"
#endif

#include "unc_id.h"
#include "unc_data.h"
#include "unc_prot.h"

static const char *compatible_vendor_names[] = {
    UNC_VENDOR_NAME,
    "UNIControllers http://www.unicontrollers.com",
    NULL
};

static const char *compatible_device_names[] = {
    UNC_DEVICE_NAME,
    UNC_DEVICE_NAME_001_LEGACY,
    NULL
};

static const char *deprecated_device_names[] = {
    UNC_DEVICE_NAME_001_DEPRECATED,
    NULL
};

#ifndef USE_HIDAPI  /* this stuff is not needed for hidapi */
/* shamelessly stolen from Christian Starkjohann's hiddata.c file */
#define USBRQ_HID_GET_REPORT    0x01
#define USBRQ_HID_SET_REPORT    0x09
#define USB_HID_REPORT_TYPE_FEATURE 3
#endif

#include "uncusb.h"

/* public (global) variable */
/* please note it is not used with hidapi */
int unc001_usb_timeout = 5000;

/*  for error handling */
static const char *last_uncusb_errorstr = NULL;


void unc001_usb_init()
{
#ifdef USE_HIDAPI
    /* no initialization required for hidapi */
#else
    usb_init();
    usb_find_busses();
    usb_find_devices();
#endif
}

void unc001_usb_rescan()
{
#ifdef USE_HIDAPI
    /* no global variables are used for hidapi; rescan is a no-op */
#else
    usb_find_devices();
#endif
}


static int hexdig2num(char c)
{
    if(c >= '0' && c <= '9') return c - '0';    
    if(c >= 'A' && c <= 'F') return c - 'A' + 10;    
    if(c >= 'a' && c <= 'f') return c - 'a' + 10;    
    return -1;
}

static int extract_sernum(const char *str, int sernum[3])
{
    int p = 0;
    int n = 0;
    sernum[0] = sernum[1] = sernum[2] = -1;
    for(; *str && p<3; str++) {
        if(*str == ':') {
            sernum[p] = n;
            p++;
            n = 0;
        } else {
            int d = hexdig2num(*str);
            if(d == -1) return 0;
            n = n*16 + d;
        }
    }
    if(!*str) {
        sernum[p] = n;
        p++;
    }
    return p==3;
}


#ifdef USE_HIDAPI
/* hidapi uses wchar_t arrays but this is senseless for 
   USB strings as they SHOULD contain ASCII chars only
   unless the vendor goes totally mad.  Hell if there
   is a point to make your device unusable for all the
   world outside of your non-latin country; be wise,
   use English.  What we definitely can say is that
   Unicontrollers' devices will ALWAYS return ASCII
   strings without non-ascii chars.
   Anyway, what we need is a convertor.
   The returned buffer is malloc'ed; make sure you
   free() it after use.
*/
static char *simple_from_wchar(const wchar_t *array)
{
   int len =0, i;
   char *res;
   while(array[len]) len++;
   res = (char*) malloc(len+1);
   for(i=0; i<len; i++)
       res[i] = array[i]>127 ? '?' : array[i];
   res[len] = 0;
   return res;
}

#endif

struct unc001_item *unc001_scan(int unc001_id)
{
#ifdef USE_HIDAPI
    struct unc001_item *res = NULL;
    struct hid_device_info *devs, *cur_dev;


    devs = hid_enumerate(0, 0);
    for(cur_dev = devs; cur_dev; cur_dev=cur_dev->next) {
        if(cur_dev->vendor_id == UNC_VENDOR_ID &&
            cur_dev->product_id == UNC_DEVICE_ID)
        {
            char *ss; /* simple string */
            const char * const *p;
            int nm, depr;

            ss = simple_from_wchar(cur_dev->manufacturer_string);
            nm = 0; depr = 0;
            for(p = compatible_vendor_names; *p; p++)
                if(0==strcmp(ss, *p))
                    nm = 1;
            free(ss);
            if(!nm) continue;

            ss = simple_from_wchar(cur_dev->product_string);
            nm = 0;
            for(p = compatible_device_names; *p; p++)
                if(0==strcmp(ss, *p))
                    nm = 1;
            for(p = deprecated_device_names; *p; p++)
                if(0==strcmp(ss, *p)) {
                    nm = 1;
                    depr = 1;
                }
            free(ss);
            if(!nm) continue;

            /**/
            {
                struct unc001_item *tmp;
                tmp = (struct unc001_item *)
                   malloc(sizeof(struct unc001_item));
                ss = simple_from_wchar(cur_dev->serial_number);
                extract_sernum(ss, tmp->sernum);
                free(ss);
                if(unc001_id == 0 || unc001_id == tmp->sernum[2]) {
                    tmp->dev_hdl = (void*)hid_open_path(cur_dev->path);
                    if(!tmp->dev_hdl) {
                        free(tmp);
                    } else {
                        tmp->next = res;
                        res = tmp;
                        if(depr) {
                            tmp->version = 0;
                            hid_close(tmp->dev_hdl);
                            tmp->dev_hdl = NULL;
                        } else {
                            struct unc001_global_configuration_str s1;
                            int r = unc001_get_configuration(tmp->dev_hdl, &s1);
                            tmp->version = (r == -1) ? 0 : s1.version;
                        }
                    }
                } else {
                    free(tmp);
                }
            }
        }
    }
    hid_free_enumeration(devs);
    return res;

#else

    struct usb_bus *bus;
    struct usb_device *dev;
    struct unc001_item *res = NULL;
    char buf[256];
    int rc;

    for(bus=usb_get_busses(); bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if((dev->descriptor.idVendor == UNC_VENDOR_ID &&
               dev->descriptor.idProduct == UNC_DEVICE_ID))
            {
                int matched = 1;
                int deprecated = 0;
                usb_dev_handle *handle = usb_open(dev);
                if(!handle) {
                    last_uncusb_errorstr = usb_strerror();
                    continue;
                }
                if(!dev->descriptor.iManufacturer) {
                    /* no manufacturer string at all */
                    matched = 0;
                } else {
                    rc = usb_get_string_simple(handle,
                                       dev->descriptor.iManufacturer,
                                       buf, sizeof(buf));
                    if(rc<0) {
                        last_uncusb_errorstr =  usb_strerror();
                    } else
                    if(rc == 0) {
                        matched = 0;
                    } else {
                        const char * const *p;
                        int nm = 0;
                        for(p = compatible_vendor_names; *p; p++)
                            if(0==strcmp(buf, *p))
                                nm = 1;
                        matched = matched && nm;
                    }
                }
                if(!dev->descriptor.iProduct) {
                    /* no product string at all */
                    matched = 0;
                } else {
                    rc = usb_get_string_simple(handle,
                                       dev->descriptor.iProduct,
                                       buf, sizeof(buf));
                    if(rc<0) {
                        last_uncusb_errorstr =  usb_strerror();
                    } else
                    if(rc == 0) {
                        matched = 0;
                    } else {
                        const char * const *p;
                        int nm = 0;
                        for(p = compatible_device_names; *p; p++)
                            if(0==strcmp(buf, *p))
                                nm = 1;
                        for(p = deprecated_device_names; *p; p++)
                            if(0==strcmp(buf, *p)) {
                                nm = 1;
                                deprecated = 1;
                            }
                        matched = matched && nm;
                    }
                }
                if(!dev->descriptor.iSerialNumber) {
                    /* no serial number string at all */
                    //matched = 0;
                    if(matched) deprecated = 1;
                } else {
                    rc = usb_get_string_simple(handle,
                                        dev->descriptor.iSerialNumber,
                                        buf, sizeof(buf));
                    if(rc<0) {
                        last_uncusb_errorstr =  usb_strerror();
                        continue;
                    }
                    if(rc == 0) {
                        //matched = 0;
                        if(matched) deprecated = 1;
                    }
                }
                // now it's the Serial in the buf
                if(matched) {
                    struct unc001_item *tmp;
                    tmp = (struct unc001_item *)
                       malloc(sizeof(struct unc001_item));
                    tmp->dev_hdl = handle;
                    extract_sernum(buf, tmp->sernum);
                    if(unc001_id == 0 || unc001_id == tmp->sernum[2]) {
                        tmp->next = res;
                        res = tmp;
                        if(deprecated) {
                            tmp->version = 0;
                            usb_close(handle);
                            tmp->dev_hdl = NULL;
                        } else {
                            struct unc001_global_configuration_str s1;
                            int r = unc001_get_configuration(handle, &s1);
                            tmp->version = (r == -1) ? 0 : s1.version;
                        }
                    } else {
                        usb_close(handle);
                        free(tmp);
                    }
                } else { /* not matched, not interesting... */
                    usb_close(handle);
                }
            }
        }
    }
    return res;
#endif
}

void unc001_destroy_list(struct unc001_item *lst)
{
    while(lst) {
        struct unc001_item *tmp = lst;
        lst=lst->next;
        if(tmp->dev_hdl)
            unc001_close(tmp->dev_hdl);
        free(tmp);
    }
}

void unc001_close(void *dev_hdl)
{
#ifdef USE_HIDAPI
    hid_close((hid_device*)(dev_hdl));            
#else
    usb_close(dev_hdl);
#endif
}

static
int get_info_from_device(void *dev_hdl, int query_id, void *buf, int len)
{
#ifdef USE_HIDAPI
  #if 0
    int res;
    unsigned char *buf2 = (unsigned char*)malloc(len+1);
    buf2[0] = query_id;
    res = hid_get_feature_report((hid_device*) dev_hdl, buf2, len+1);
    //res = HidD_GetFeature(((hid_device*)dev)->device_handle, buf3, len+1);
    memcpy(buf, buf2, len);
    free(buf2);
    return res==-1 ? -1 : len;
  #else
    ((char*)buf)[0] = query_id;
    return hid_get_feature_report((hid_device*) dev_hdl, buf, len);
  #endif
#else
    return
        usb_control_msg(dev_hdl,
                        USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                        USBRQ_HID_GET_REPORT,
                        USB_HID_REPORT_TYPE_FEATURE << 8 |
                            (query_id & 0xff),
                        0, buf, len, unc001_usb_timeout);
#endif
}

static
int send_info_to_device(void *dev_hdl, const void *buf, int len)
{
#ifdef USE_HIDAPI
    return hid_send_feature_report((hid_device*) dev_hdl, buf, len);
#else
    return
        usb_control_msg(dev_hdl,
                        USB_TYPE_CLASS | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                        USBRQ_HID_SET_REPORT,
                        USB_HID_REPORT_TYPE_FEATURE << 8 |
                            (*(const char*)buf & 0xff),
                        0, (void*)buf, len, unc001_usb_timeout);
#endif
}

static
int send_command(void *dev_hdl, int cmd, int param, int p2, uint32_t data)
{
    struct {
        uint8_t req_num, cmd, param;
        uint8_t param2;
        uint32_t data;
    } buf;
    int r;

    buf.req_num = uncreq_command;
    buf.cmd = cmd;
    buf.param = param;
    buf.param2 = p2;
    buf.data = data;
    r = send_info_to_device(dev_hdl,  &buf, sizeof(buf));
    if(r != sizeof(buf)) {
        last_uncusb_errorstr = "error sending info to device (wrong count)";
        return -1;
    }
    return 0;
}

#if 0
int unc001_query_state(void *dev_hdl,
                       int *version,
                       int *state,
                       int counters[UNC_INPUTS_COUNT],
                       unsigned long reactions[UNC_INPUTS_COUNT])
{
    unc001_global_configuration_str conf;
    unc001_global_state_str statstr;
    int r;
    int i;

    r = unc001_get_fullstate(dev_hdl, &statstr);
    if(r == -1) return -1;

    r = unc001_get_configuration(dev_hdl, &conf);
    if(r == -1) return -1;

    if(version) *version = conf.version;
    if(state) *state = statstr.state_bits;
    for(i=0; i<UNC_INPUTS_COUNT; i++) {
         if(counters) counters[i] = statstr.counters[i];
         if(reactions) reactions[i] = conf.reactions[i];
    }

    return 0;
}
#endif

static int unc001_get_structure(void *dev_hdl, int req, void *data, int sz)
{
    int r;
    r = get_info_from_device(dev_hdl, req, data, sz);
    if(r <= 0) {
#ifdef USE_HIDAPI
        last_uncusb_errorstr = "get_info_from_device failed";
#else
        last_uncusb_errorstr = usb_strerror();
#endif
        return -1;
    }
    if(r != sz) {
        last_uncusb_errorstr = "wrong count on reading from the device";
        return -1;
    }
    return 0;
}

int unc001_get_fullstate(void *dev_hdl, struct unc001_global_state_str *data)
{
    return unc001_get_structure(dev_hdl, uncreq_query_fullstate,
                                data, sizeof(*data));
}

int unc001_get_extrastate(void *dev_hdl, struct unc001_extra_state_str *data)
{
    return unc001_get_structure(dev_hdl, uncreq_query_extrastate,
                                data, sizeof(*data));
}

int unc001_get_configuration(void *dev_hdl,
                             struct unc001_global_configuration_str *data)
{
    return unc001_get_structure(dev_hdl, uncreq_query_configuration,
                                data, sizeof(*data));
}

int unc001_save_to_eeprom(void *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_save_to_eeprom, 0, 0, 0);
}
int unc001_restore_from_eeprom(void *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_restore_from_eeprom, 0, 0, 0);
}
int unc001_restore_factory(void *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_restore_factory, 0, 0, 0);
}
int unc001_clear_msg(void *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_clear_msg, 0, 0, 0);
}

int unc001_set_new_id(void *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_id, 0, 0, id);
}

int unc001_set_configuration(void *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_userconfig, 0, 0, id);
}

/* use with EXTREME CAUTION!!! better don't use at all */
int unc001_set_hardware_conf(void *dev_hdl, unsigned int id)
{
    return send_command(dev_hdl, uncreqcmd_set_hwconfig, 0, 0, id);
}

int unc001_perform_action(void *dev_hdl, int action)
{
    return send_command(dev_hdl, uncreqcmd_action, 0, 0, action);
}

int unc001_set_reactions(void *dev_hdl, int line, unsigned long reaction)
{
    return send_command(dev_hdl, uncreqcmd_set_reactions, line, 0, reaction);
}

int unc001_get_counter_range(void *dev_hdl, int *min, int *max)
{
    struct unc001_global_configuration_str conf;
    int r = unc001_get_configuration(dev_hdl, &conf);
    if(r>=0) {
        *min = ((conf.hardware_config & 0x0f) == 0x04) ? 0 : 4;
        *max = ((conf.hardware_config & 0xf0) == 0x40) ? 7 : 3;
    }
    return r;
}

int unc001_zero_query_counters(void *dev_hdl, int counters[], int ncount)
{
    uint8_t data[1+2*UNC_INPUTS_COUNT];
    int r, i, min=0, max=-1;
    unc001_get_counter_range(dev_hdl, &min, &max);
    r = get_info_from_device(dev_hdl,
                             uncreq_zero_query_counters,
                             data, sizeof(data));
    if(r != sizeof(data))
        return -1;
    for(i=0; i<ncount; i++) {
        int i2 = i + min;
        if(i2 > max) break;
        counters[i] = data[2*i2+1] | ((uint16_t)(data[2*i2+2])<<8) ;
    }
    return i;
}

int unc001_display_message(void *dev_hdl, const char *msg)
{
    char buf[16]; // must be 16 as the request size is 16
    int sent = 0;
    int len = strlen(msg);
    int remain = len;
    while(remain > 0) {
        int k;
        int i;
        buf[0] = uncreq_long_command;
        buf[1] = uncreqlcmd_com_message;
        for(i = 2; i<sizeof(buf); i++)
            buf[i] = i+sent-2<len ? msg[i+sent-2] : 0; 
        k = send_info_to_device(dev_hdl, buf, sizeof(buf));
        if(k >= 0) {
            sent += (sizeof(buf)-2);
            remain -= (sizeof(buf)-2);
        } else
            break;
    }
    return (sent >= len) ? 0 : -1;
}

int unc001_com_poll(void *dev_hdl, char *msg, int len)
{
    int got = 0;
    memset(msg, 0, len);    
    while(len-1-got >= 15) {
        int r, i;
        char buf[16];
        memset(buf, 0, sizeof(buf));    
        r = get_info_from_device(dev_hdl, uncreq_com_poll, buf, 16);
        if (r<0 || !buf[1]) break;
        for(i=1; i<sizeof(buf); i++)
            if((msg[got]=buf[i]))
                got++;
            else
                break;
    }
    return got;
}

int unc001_com_release(void *dev_hdl)
{
    return send_command(dev_hdl, uncreqcmd_com_release, 0, 0, 0);
}

int unc001_dump_eeprom(void *dev_hdl, void* buf)
{
    return get_info_from_device(dev_hdl, uncreq_query_eeprom, buf, 128);
}

int unc001_takeover_stream(void *dev_hdl)
{
#ifdef USE_HIDAPI
    /* hopefully nothing to be done... */
    return 0;
#else
    if(LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP)
    {
        // Detach kernel driver (usbhid) from device interface. (Linux hack)
        usb_detach_kernel_driver_np(dev_hdl, 0);
        usb_detach_kernel_driver_np(dev_hdl, 1);
    }
    if(usb_set_configuration(dev_hdl, 1) < 0){
        last_uncusb_errorstr =  usb_strerror();
        return -1;
    }
    if(usb_claim_interface(dev_hdl, 0) < 0){
        last_uncusb_errorstr =  usb_strerror();
        return -1;
    }
    return 0;
#endif
}

int unc001_read_stream(void *dev_hdl, char buf[8])
{
#ifdef USE_HIDAPI
    return hid_read(dev_hdl, buf, 8);
#else
    return usb_interrupt_read(dev_hdl, 0x81, buf, 8, 0);
#endif
}

static
int unc001_i2c_command(void *dev_hdl, int i2c_la, const void *msg, int len)
{
    char buf[16]; // must be 16 as the request size is 16
    int i, k;
    int maxlen = ((i2c_la & 1) == 1) ? 7 : 12;
    buf[0] = uncreq_long_command;
    buf[1] = uncreqlcmd_bus_command;
    buf[2] = i2c_la;
    buf[3] = len > maxlen ? maxlen : len;
    if(msg) {
        for(i = 4; i<sizeof(buf); i++)
            buf[i] = (i - 4 < len) ? ((char*)msg)[i - 4] : 0; 
    }
    k = send_info_to_device(dev_hdl, buf, sizeof(buf));
    return (k >= 0) ? 0 : -1;
}

int unc001_i2c_write(void *dev_hdl, int i2c_addr, const void *buf, int blen)
{
    int la = (i2c_addr << 1);
    return unc001_i2c_command(dev_hdl, la, buf, blen);
}

int unc001_i2c_read(void *dev_hdl, int i2c_addr, int len)
{
    int la = (i2c_addr << 1) | 1; /* 1 is the reading direction flag */
    return unc001_i2c_command(dev_hdl, la, NULL, len);
}

int unc001_write_e_reg(void *dev_hdl, int regn, char data[8])
{
    char buf[16]; // must be 16 as the request size is 16
    int k;
    int i;
    buf[0] = uncreq_long_command;
    buf[1] = uncreqlcmd_write_e_reg;
    buf[2] = regn;
    for(i = 0; i < 8; i++)
        buf[i+3] = data[i]; 
    k = send_info_to_device(dev_hdl, buf, sizeof(buf));
    return k==16 ? 0 : -1;
}





const char *unc001_strerror()
{
    return last_uncusb_errorstr;
}



/*
 * These functions are not static but intentionally not included into 
 * the header because they are NOT intended for normal use; the only
 * program which uses them is wrongreq.c; don't use these functions
 * unless you're absolutely sure you know what you're doing.
 */
int unc001_simple_request(void *dev_hdl, int req)
{
    return send_command(dev_hdl, req, 0, 0, 0);
}
int unc001_raw_data_request(void *dev_hdl, int query_id, void *buf, int len)
{
    return get_info_from_device(dev_hdl, query_id, buf, len);
}

