/*
 * UNC001 host-side software. (c) Unicontrollers, Ltd., 2010-2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file uncusb.h
 * Headers of public functions for communicating with the device over USB
 *
 * Written by Andrey V. Stolyarov
 *
 */


#ifndef UNCUSB_H_SENTRY
#define UNCUSB_H_SENTRY

#ifndef UNC_INPUTS_COUNT
#define UNC_INPUTS_COUNT 8
#endif

#include <stdint.h>


#ifdef __cplusplus
extern "C" {
#endif


    /* This function must be called before any other functions will */
    /* Actually, it initializes global data of the libusb library,
     * so in case your application uses libusb for communicating
     * devices other than UNC001, you should only initialize it once,
     * but make sure the initialization happens, one way or another.
     * If you only use USB to communicate UNC001 within your program,
     * that is, you don't call libusb directly, then just ignore all
     * this stuff and simply call this function.
     */
void unc001_usb_init();

    /* Call this function to rescan the bus for connected devices */
void unc001_usb_rescan();

    /* This structure is used to form a simple unidirection list to
       store information about all the UNC001 devices found on your bus.
     */
struct unc001_item {
    int sernum[3];
    int version;      /* 0 for devices with deprecated protocols */
    void *dev_hdl;    /* NULL for deprecated/incompatible devices */
    struct unc001_item *next;
};

    /* the list MUST be destroyed later with unc001_destroy_list */
struct unc001_item *unc001_scan(int unc001_id);
    /* all dev_hdl's are closed; to avoid this, assign NULL to them */
void unc001_destroy_list(struct unc001_item *lst);

void unc001_close(void *dev_hdl);

#if 0
int unc001_query_state(void *dev_hdl,
                       int *version,
                       int *state,
                       int counters[UNC_INPUTS_COUNT],
                       unsigned long reactions[UNC_INPUTS_COUNT]);
#endif


struct unc001_global_configuration_str;
int unc001_get_configuration(void *dev_hdl,
                             struct unc001_global_configuration_str *data);

struct unc001_global_state_str;
int unc001_get_fullstate(void *dev_hdl,
                         struct unc001_global_state_str *data);

struct unc001_extra_state_str;
int unc001_get_extrastate(void *dev_hdl,
                          struct unc001_extra_state_str *data);


int unc001_save_to_eeprom(void *dev_hdl);
int unc001_restore_from_eeprom(void *dev_hdl);
int unc001_restore_factory(void *dev_hdl);

int unc001_clear_msg(void *dev_hdl);

int unc001_set_new_id(void *dev_hdl, unsigned int id);
int unc001_set_configuration(void *dev_hdl, unsigned int id);
    /* use with EXTREME CAUTION!!! better don't use at all */
int unc001_set_hardware_conf(void *dev_hdl, unsigned int id);

int unc001_perform_action(void *dev_hdl, int action);
int unc001_set_reactions(void *dev_hdl, int line, unsigned long reaction);
int unc001_get_counter_range(void *dev_hdl, int *min, int *max);
int unc001_zero_query_counters(void *dev_hdl, int counters[], int ncount);

    /* buf must point to a buffer of 128 bytes */
int unc001_dump_eeprom(void *dev_hdl, void* buf);


#define UNCUSB_POLL_BUF_MIN_SIZE 17

    /* sends the given string to the device's COM port */
int unc001_display_message(void *dev_hdl, const char *msg);
    /* grabs the device's COM port and polls it for chars;
     * chars are written to the buffer, at most len-1, and
     * adds the termination zero thus making a correct string.
     * len MUST be at least UNCUSB_POLL_BUF_MIN_SIZE bytes,
     * or else the function will not work; this is because the
     * request is 16 bytes long, and there's no storage to keep
     * the rest of it in case there's no room in the msg.
     */
int unc001_com_poll(void *dev_hdl, char *msg, int len);
    /* Tells the device that COM port input should
     * again be interpreted as AT commands
     */
int unc001_com_release(void *dev_hdl);

    /* Prepares to read the "interrupt input" from the device.
     * This requires to tell the system driver that the stream
     * is no longer to be handled by the system.
     */
int unc001_takeover_stream(void *dev_hdl);
    /* Tries to read the "interrupt input" from the device.
     * The buffer must be exactly 8 bytes long!
     * Function returns the actual number of received bytes,
     * or -1 on error
     */
int unc001_read_stream(void *dev_hdl, char buf[8]);

    /* Tries to send info via the i2c bus */
int unc001_i2c_write(void *dev_hdl, int i2c_addr, const void *buf, int blen);
    /* Requests information on the i2c bus */
int unc001_i2c_read(void *dev_hdl, int i2c_addr, int len);

    /* Overwrites the given Exx register */
int unc001_write_e_reg(void *dev_hdl, int regn, char data[8]);

extern int unc001_usb_timeout;

const char *unc001_strerror();

#ifdef __cplusplus
}
#endif

#endif
