/*
 * UNC001 firmware. (c) Unicontrollers, Ltd., 2010-2011
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file atparser.c
 * Implements the finite state machine used to analyse AT commands
 * on the COM port.  If your device has no COM port, ignore this module.
 *
 * Written by Andrey V. Stolyarov
 *
 */

#include "options.h"
#if UNC001_ENABLE_AT_COMMANDS == 1

#include <avr/pgmspace.h>

#include "ser_io.h"
#include "print.h"
#include "unc_prim.h"
#include "misc.h"

#include "atparser.h"


static int8_t state = 0; /* 0 for home, >0 for current pos., -1 for err */

static void action_ok()
{
    print_str("Ok\n");
}

static void action_error()
{
    print_str(PSTR("ERROR @"));
    print_int(state);
    print_char('\n');
}

/* ATH0 ATH1 */
static void action_ath(int n)
{
    perform_action(n == 0 ? 2 : 1);
}


#include "usbdrv.h"  /* for vendor/device strings */
/* ATI ATI0 ATI1 ATI9 */
static void action_ati(int n)
{
    int *p = (n<2) ? usbDescriptorStringVendor : usbDescriptorStringDevice;
    uint8_t i;
    for(i=0;i<(n<2?USB_CFG_VENDOR_NAME_LEN:USB_CFG_DEVICE_NAME_LEN);++i)
        print_char(pgm_read_word(++p)); /* ++p is not a mistake! */
    print_char('\n');
}

/* AT%T -- print hexadecimal "all lines state" register */
void action_statequery()
{
    sync_global_state();
    print_hex(global_state.state_bits);
    print_char('\n');
}

/* AT%Z, AT%P */
void action_query_and_zero_counters()
{
    uint16_t c4[8];
    uint8_t i;
    zero_counters(c4);
    for(i=0; i<UNC_INPUTS_COUNT; i++) {
        print_int(c4[0]);
        print_char(' ');
    }
    print_char('\n');
}

/* *********************************************** */ 
/* action_action */

/*#define COM_ACTION_NUMERIC*/ /* this costs 200+ bytes */

#ifdef COM_ACTION_NUMERIC
static void action_action_bin(char *databuf)
{
    int action = 0;
    while(*databuf) {
        action *= 2;
        switch(*databuf) {
        case '1':
            action += 1;
            /* intentionally no break here */
        case '0':
            break;
        default:
            action_error();
            return;
        }
        databuf++;
    }
    perform_action(action & 0xff);
}

static void action_action_hex(char *databuf)
{
    int8_t d1, d2;
    d1 = hexdig2num(databuf[0]);
    if(d1 != -1) {
        d2 = hexdig2num(databuf[1]);
        if(d2 != -1 && databuf[2] == '\0') {
            perform_action(d1<<4 | d2);
            return;
        }
    }
    action_error();
}
#endif

#ifdef COM_ACTION_NUMERIC
static void action_action_mnem(char *databuf)
#else
static void action_action(char *databuf)
#endif
{
    int action = 0;
    while(*databuf) {
        action <<= 2;
        switch(*databuf) {
        case '=':
            break;       /* 00 == keep untouched */
        case '^':
            action |= 3; /* 11 == switch state */
            break;
        case '+':
            action |= 1; /* 01 == set on */
            break;
        case '-':
            action |= 2; /* 10 == set off */
            break;
        default:
            action_error();
            return;
        }
        databuf++;
    }
    perform_action(action & 0xff);
}

#ifdef COM_ACTION_NUMERIC
static void action_action(char *databuf)
{
    switch(databuf[0]) {
        case '0':
        case '1':
            action_action_bin(databuf);
            break;
        case 'x':
        case 'X':
            action_action_hex(databuf+1);
            break;
        case '+':
        case '-':
        case '^':
        case '=':
            action_action_mnem(databuf);
            break;
        default:
            action_error();
    }
}
#endif
/* ----------------- the parser ------------------ */

static char databuf[13];
static const char atseq[2] = { 'A', 'T' };


static void at_finalize_percent()
{
    switch(databuf[1]) {
        case 'T':
            if(state == 4) {
                action_statequery();
                return;
            }
        case 'Z':
        case 'P':
            if(state == 4) {
                action_query_and_zero_counters(databuf[1]=='P');
                return;
            }
        case 'H':
            databuf[state-2] = 0;
            action_action(databuf+2);
            return;
    }
    action_error();
}

static void at_finalize() {
    ++force_to_serial;
    if(state<3) {
        action_ok();
        goto quit;
    }
    switch(databuf[0]) {
        case 'H':
            if(state == 4 && databuf[1] >= '0' && databuf[1] <= '1') {
                action_ath(databuf[1] - '0');
                goto quit;
            }
            break;
        case 'I':
            if(state == 4 && databuf[1] >= '0' && databuf[1] <= '9') {
                action_ati(databuf[1] - '0');
                goto quit;
            }
            break;
        case '%':
            if(state > 3) {
                at_finalize_percent();
                goto quit;
            }
    }
    action_error();
quit:
    --force_to_serial;
}


/* PUBLIC FUNCTIONS */

void atp_reset()
{
    state = 0;
}

void atp_handle_char(char c)
{
    if(c>='a'&&c<='z') c -= ('a'-'A');
    switch(state) {
        case -1:
            if(c=='\n') state = 0;
            break;
        case 0:
        case 1:
            state = (c == atseq[state]) ? state+1 : 0;
            break;
        default:
            if(c == '\n') {
                at_finalize();
                state = 0;
            } else
            if(state-2 > sizeof(databuf)-1) {
                state = -1;
            } else {
                databuf[state++ - 2] = c;
            }
    }
}

#endif
