#if UNC001_ENABLE_I2C_SUBSYSTEM > 0

#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "print.h"
#include "unc_prim.h"
#include "twowire.h"

/* tunable parameters */

#define MAX_PAYLOAD 12

#define IN_BUF_LENGTH 7 
#define CYCLES_FOR_TIMEOUT 10

/* end of tunable params */


#define slaveF_SCL 400000 /* 400 Khz */

#define WRITEFLAG 0  /* writing */
#define READFLAG 1   /* reading */

#define TW_START                    0x08
#define TW_REP_START                0x10
#define TW_MT_SLA_ACK               0x18
#define TW_MT_DATA_ACK              0x28
#define TW_MR_SLA_ACK               0x40
#define TW_MR_DATA_ACK              0x50
#define TW_MR_DATA_NACK             0x58

static uint8_t msg_len;
static uint8_t msg_addr;
static uint8_t outbuf[MAX_PAYLOAD];

static uint8_t current_byte;

// XXXXXXXXXXXX remove this later!
static uint8_t interrupt_counter = 0;

enum states {
    st_idle,
    st_startwait,
    st_sending_addr,
    st_sending_byte,
    st_stopping,
    st_reading_byte,
    st_reading_last_byte,
};

static enum states state = st_idle;

static uint8_t timemark;

static void try_sending()
{
    if(state != st_idle) return;
    TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWSTA)|(1<<TWEN);
    state = st_startwait;
}

static void abort_msg()
{
    state = st_idle;    
}

void read_next_byte()
{
    if(current_byte + 1 < msg_len) {
        TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWEN)|(1<<TWEA); 
        state = st_reading_byte;
    } else {
        TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWEN);
        state = st_reading_last_byte;
    }
}

static void i2c_err(uint8_t code, uint8_t byte)
{
    print_str("TW:");
    print_char(code);
    print_hex_byte(byte);
    // XXXXXXXXXX
    print_hex_byte(msg_addr);
    print_hex_byte(interrupt_counter);
    print_hex_byte(state);
    interrupt_counter = 0;
    print_char('\n');
}

ISR(TWI_vect)
{
    uint8_t cmd = TWSR & 0xF8;
    timemark = CYCLES_FOR_TIMEOUT; 
    print_char(';');
    print_hex_byte(cmd);
    print_char('0'+state);
    if(++interrupt_counter > 5) {
        TWCR &= ~(1<<TWIE);
    }
    switch(cmd)
    {
        case TW_START:       /* 08 */
        case TW_REP_START:   /* 10 */
            TWDR = msg_addr;
            TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWEN);
            state = st_sending_addr;
#if 0
            print_char('^');
            print_hex_byte(outbuf[current_msg][1]);
            print_char('\n');
#endif
            break;
        case TW_MT_SLA_ACK:  /* 18 */
            current_byte = 0;
            /* no break here */
        case TW_MT_DATA_ACK:
            if(current_byte >= msg_len) {
                state = st_stopping;
                TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
                while(TWCR & (1<<TWSTO));
                abort_msg(); // XXXXXXX this is no abort, this is success!!!
            } else {
                state = st_sending_byte;
                print_char('b');
                TWDR = outbuf[current_byte];
                current_byte++;
                TWCR = (1<<TWIE)|(1<<TWINT)|(1<<TWEN);
            }
            break;
        case TW_MR_SLA_ACK:
            current_byte = 0;                
            read_next_byte();
            break;
        case TW_MR_DATA_ACK:
            if(current_byte < IN_BUF_LENGTH)
                extra_state.data_buffer[current_byte+1] = TWDR;
            current_byte++;
            read_next_byte();
            break;
        case TW_MR_DATA_NACK:
            if(current_byte < IN_BUF_LENGTH)
                extra_state.data_buffer[current_byte+1] = TWDR;
            extra_state.data_buffer[0] = current_byte+1;
            state = st_stopping;
            TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);
            while(TWCR & (1<<TWSTO));
            abort_msg();
            break;
        default:
            TWCR &= ~(1<<TWIE);
            i2c_err('E', cmd);
            state = st_idle;
    }
}

/*
From http://nagits.wordpress.com/2010/12/18/avr_i2c_eeprom/
*/
void tw_init()
{
    TWBR = (F_CPU/slaveF_SCL - 16)/(2 * /* TWI_Prescaler= 4^TWPS */1);
    
    if(TWBR < 10)
        TWBR = 10;
 
    TWSR &= (~((1<<TWPS1)|(1<<TWPS0)));
    TWCR |= (1<<TWIE);
}

void tw_hook()
{
    if(state == st_idle)
        return;
    if(!--timemark) {
  //    i2c_err('T', TWSR & 0xF8);
        print_char('T');
        if((msg_addr & 1) == READFLAG)
            extra_state.data_buffer[0] = 0;
        abort_msg();
    }
}

uint8_t tw_send(uint8_t addr, char *str, uint8_t len)
{
    uint8_t i;

    print_char('&');
    print_hex_byte(addr);
    print_char('0'+len);
    print_char('\n');

    if(state != st_idle) return 0;
    current_byte = 0;

    msg_len = len;
    msg_addr = (addr << 1) | WRITEFLAG;
    for(i = 0; i < len; i++)
        outbuf[i] = str[i];    

    try_sending();
    print_char('&');
    print_char('\n');
    return 1;
}

uint8_t tw_recv(uint8_t addr, uint8_t len)
{
    if(state != st_idle) return 0;

    msg_len = len;
    msg_addr = (addr << 1) | READFLAG;
    extra_state.data_buffer[0] &= (uint8_t)~7;

    try_sending();
    return 1;
}

#endif
