/*
 * UNC001 firmware. (c) Unicontrollers, Ltd., 2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc_1w.c
 * Implemetation of the 1wire(R) support
 *
 * Written by Andrey V. Stolyarov, except for the crc8() function;
 * see the comment right after that function about its status.
 *
 */

//#include <avr/wdt.h>
#include <avr/eeprom.h>

#include "print.h"
#include "unc_prim.h"

#include "onewire.h"

#include "unc_1w.h"


#define DROP_DELAY_CYCLES 25

/* DS18x20-related defines */
#define DS18S20_ID           0x10
#define DS18B20_ID           0x28
#define DS18X20_CONVERT_T    0x44
#define DS18X20_READ         0xBE

static uint8_t id_tmarks[4] = { 0, 0, 0, 0 };

static void print_id(uint8_t charcode, uint8_t checkmask, uint8_t *id)
{
    if(global_configuration.user_config_bytes[2] & checkmask) {
        int8_t j;
        print_char('W');
        print_char(charcode);
        for(j=7; j>=0; j--) print_hex_byte(id[j]);
        print_char('\n');
    }
}

static uint8_t e_reg_matches_id(uint8_t reg, const uint8_t *id, uint8_t k)
{
    for(; k<8; k++) {
        uint8_t b = eeprom_read_byte((uint8_t*) 128 + reg*8 + k);
        if(b != id[k]) return 0;
    }
    return 1;
}

static void handle_connected_ibutton(const uint8_t *id)
{
    if(global_configuration.user_config_bytes[2] & 0x08) {
        int i;
        for(i=1; i<48; i++) {
            uint8_t b = eeprom_read_byte((uint8_t*) 128 + i*8);
            if(b == 0x01) {  /* iButton id stored */
                if(e_reg_matches_id(i, id, 1)) {
                    perform_action(0xf0, 0x00);
                }
            } else {
                /* stored ids must occupy a continuous area */
                /* starting with the E01 register */
                /* if this register doesn't comply, no further will */
                return;
            }
        }
    }
}

static void handle_temp_change(int16_t last, int16_t next, uint8_t *id)
{
    if(global_configuration.user_config_bytes[2] & 0x04) {
        uint8_t k;
        for(k = 46; k >= 40; k -= 2) {
            if(e_reg_matches_id(k, id, 0)) {
                static struct slave_reg {
                    uint8_t ff;
                    uint8_t lsb_a_hi_wm;
                    uint8_t lsb_a_lo_wm;
                    uint8_t msb_a;
                    int16_t hi_wm, lo_wm;
                } s;
                eeprom_read_block(&s, (uint8_t*)128+(k+1)*8, 8);
                if(last > s.lo_wm && next <= s.lo_wm)
                    perform_action(s.msb_a, s.lsb_a_lo_wm);
                else
                if(last < s.hi_wm && next >= s.hi_wm)
                    perform_action(s.msb_a, s.lsb_a_hi_wm);
                break;
            }
        }
    }
}

static void register_id(uint8_t id[8])
{
    int8_t free_pos = -1, found_pos = -1;
    int8_t i, j;
    for(i=3; i>=0; i--) {
        if(extra_state.ow_ids[i][0] == 0) {
            /* empty slot */
            free_pos = i;
        } else {
            for(j=0; j<8; j++)
                if(id[j] != extra_state.ow_ids[i][j])
                    goto no_match;
            /* here is it! */
            found_pos = i;
            break;
        no_match:
            ;
        }
    } 
    if(found_pos == -1) { /* not found => new connected device */
        print_id((free_pos == -1) ? '#' : '+', 0x40, id);
        if(free_pos == -1) /* no free slots -> ignoring silently */
            return;
        for(j=0; j<8; j++)
            extra_state.ow_ids[free_pos][j] = id[j];
        found_pos = free_pos;
        if(id[0] == 0x01) { /* iButton "key" */
            handle_connected_ibutton(id);
        }
    }
    id_tmarks[found_pos] = global_state.timemark_bytes[0];
}

static void drop_absent_ids()
{
    int8_t i;
    for(i=0; i<4; i++) {
        if(extra_state.ow_ids[i][0] != 0 &&
           (uint8_t)(global_state.timemark_bytes[0] - id_tmarks[i])
               > DROP_DELAY_CYCLES
        ) {
            print_id('-', 0x20, extra_state.ow_ids[i]);
            extra_state.ow_ids[i][0] = 0;
        }
    }
}


static uint8_t crc8(uint8_t *data_in, uint8_t data_len)
{
    enum { CRC8INIT = 0x00, CRC8POLY = 0x18 }; /* 0X18 = X^8+X^5+X^4+X^0 */
    uint8_t  crc;
    uint8_t  loop_count;
    uint8_t  bit_counter;
    uint8_t  data;
    uint8_t  feedback_bit;
    crc = CRC8INIT;
    for(loop_count = 0; loop_count != data_len; loop_count++)
    {
        data = data_in[loop_count];
        bit_counter = 8;
        do {
            feedback_bit = (crc ^ data) & 0x01;
            if ( feedback_bit == 0x01 ) {
                crc = crc ^ CRC8POLY;
            }
            crc = (crc >> 1) & 0x7F;
            if ( feedback_bit == 0x01 ) {
                crc = crc | 0x80;
            }
            data = data >> 1;
            bit_counter--;
        } while (bit_counter > 0);
    }
    return crc;
}
/*
The function crc8() code is based on the one by:
    Colin O'Flynn - Copyright (c) 2002 
    with "only minor changes" by M.Thomas 9/2004

    Modified by Andrey V. Stolyarov for Unicontrollers, Ltd., to comply
    the Unicontrollers' coding conventions.

The following permission applies to this particular function only:

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.
*/


static void scan_temperature()
{
    if(global_configuration.user_config_bytes[2] & 0x10) {
        static uint8_t counter = 0;
        uint8_t i;
        ++counter;
        if(counter > 50)
            counter = 0;
        else
        if(counter == 3) {
            /* scan for the sensors and tell them to calc. temperatures */
            for(i=0; i<4; i++)
                if(extra_state.ow_ids[i][0] == DS18S20_ID ||
                   extra_state.ow_ids[i][0] == DS18B20_ID
                ) {
                    ow_command(DS18X20_CONVERT_T, extra_state.ow_ids[i]);
                }
        } else
        if(counter == 10) {
            /* read the temperatures */
            for(i=0; i<4; i++) {
                if(extra_state.ow_ids[i][0] == DS18S20_ID ||
                   extra_state.ow_ids[i][0] == DS18B20_ID
                ) {
                    uint8_t buf[9];
                    uint8_t j;
                    ow_command(DS18X20_READ, extra_state.ow_ids[i]);
                    for(j=0; j<9; j++)
                        buf[j] = ow_byte_rd();
                    if(crc8(buf, 8) == buf[8]) {
                        int16_t t = buf[1];
                        t <<= 8;
                        t |= buf[0];
                        if(extra_state.ow_ids[i][0] == DS18S20_ID)
                            t <<= 3;
                        if((uint16_t)(extra_state.ow_temperature[i]) !=
                           (uint16_t)0x8000 &&
                           extra_state.ow_temperature[i] < 0x0540)
                        {
                            handle_temp_change(
                                extra_state.ow_temperature[i], t,
                                extra_state.ow_ids[i]);
                        }
                        extra_state.ow_temperature[i] = t;
                    }
                } else {
                    extra_state.ow_temperature[i] = 0x8000;
                }
            }
        }
    }
}

void unc1w_init()
{
    uint8_t k;
    ow_set_bus(&PINB, &PORTB, &DDRB, PB0);
    for(k = 0; k < 4; k++) {
        extra_state.ow_ids[k][0] = 0;
        extra_state.ow_temperature[k] = 0x8000;
    }
}

void unc1w_hook()
{
    uint8_t id[8], diff, rr;
    if(!(global_configuration.user_config_bytes[2] & 0x80)) return;
    if(!(global_state.timemark_bytes[0] & 0x0f)) { /* once in 16 ticks */
        rr = ow_reset();
        if(rr == 0) {
            for(diff = OW_SEARCH_FIRST; diff != OW_LAST_DEVICE;)
            {
                diff = ow_rom_search(diff, id);
                if(diff == OW_PRESENCE_ERR || diff == OW_DATA_ERR)
                    return; /* don't perform anything else on this tick */
                register_id(id);
            }
        }
        drop_absent_ids();
    }
    scan_temperature();
}
