/*
 * UNC001 firmware. (c) Unicontrollers, Ltd., 2010-2012
 *
 * Please see the files README, COPYING, LICENSE and WARRANTY
 * for some useful details on the legal status of this code.
 * Visit the site http://www.unicontrollers.com for more info.
 *
 * file unc_prim.c
 * Primary functions of the device: what the device actually can do
 *
 * Written by Andrey V. Stolyarov & Svyatoslav Tolchenov
 *
 */

#include <stdint.h>
#include <stddef.h>
#include <avr/eeprom.h>

#include "ucd.h"
//#include "led.h"

#include "misc.h"
#include "print.h"

#include "unc_prim.h"

#define MAX_DELAYED_ACTIONS 4

static uint32_t delayed_action_counters[MAX_DELAYED_ACTIONS];
static struct { uint8_t grp, act; } delayed_actions[MAX_DELAYED_ACTIONS];


static uint16_t input_timestamps[UNC_INPUTS_COUNT] =
    { 0, 0, 0, 0, 0, 0, 0, 0 };
uint8_t confirmed_input_state = 0;

struct unc001_global_configuration_str global_configuration;
struct unc001_global_state_str global_state;
struct unc001_extra_state_str extra_state;

#define GSTMRK (global_state.timemark_words[0])

static uint8_t get_extension(uint8_t n)
{
    return (global_configuration.hardware_config_bytes[0] >> (4*n)) & 0x0f;
}

uint8_t get_actual_input_state_byte()
{
    uint8_t res = 0;
    /*if(get_extension(1) == 0x04) {*/
    if(get_extension(1)) {
        uint8_t i20=(~IPORT2&(1<<IPORT2_IN0))>>IPORT2_IN0;
        uint8_t i21=(~IPORT2&(1<<IPORT2_IN1))>>IPORT2_IN1;
        uint8_t i22=(~IPORT2&(1<<IPORT2_IN2))>>IPORT2_IN2;
        uint8_t i23=(~IPORT2&(1<<IPORT2_IN3))>>IPORT2_IN3;
        res |= ((i20<<4) | (i21<<5) | (i22<<6) | (i23<<7));
    }
    /*if(get_extension(0) == 0x04) {*/
    if(get_extension(0)) {
        res |= (~IPORT) & 0x0f;
    }
    return res;
}

static void cancel_delayed_actions(uint8_t ac)
{
    uint8_t i;
    for(i=0; i<MAX_DELAYED_ACTIONS; i++)
        if(ac == 0xff ||
            (delayed_actions[i].grp == 0xf0 && delayed_actions[i].act == ac))
        {
            delayed_actions[i].grp = 0xff;
        }
}

void init_primary_hardware()
{
    eeprom_read_block(&global_configuration.hardware_config,
                      (unsigned char *)0 + 16, 48);
    zero_counters(NULL);
    global_state.state_bits_bytes[0] = 0;
    global_state.state_bits_bytes[1] = get_actual_input_state_byte();
    global_state_drop_message();

    if(get_extension(0) == 0x0f) {
        /* the first ext. is output */
        DDRC=0x0F;
        PORTC=0x00;
    } else {
        DDRC=0;
        PORTC=0x0F;
    }
    if(get_extension(1) == 0x0f) {
        /* the second ext. is output */
        DDRB|=0x3a;
        PORTB&=~0x3a;
    } else {
        DDRB&=~0x3a;
        PORTB|=0x3a;
    }

    DDRD|=0xF0;
    PORTD&=0x0F;

    confirmed_input_state = get_actual_input_state_byte();

    cancel_delayed_actions(0xff /* all */);
}

void restore_factory_defaults()
{
    uint8_t i;
    for(i=0; i<UNC_INPUTS_COUNT; i++)
        global_configuration.reactions[i] = 0x60000000;
    global_configuration.user_config = 0x00000014;
    zero_counters(NULL);
    global_state.state_bits = 0;
}

void global_state_putchar(uint8_t ch)
{
    global_state.message_buffer[global_state.buf_end] = ch;
    ++global_state.buf_end;
    global_state.buf_end &= 0x1f;
    if(global_state.buf_end == global_state.buf_start) {
        global_state.message_buffer[(global_state.buf_start+1)&0x1f] = '>';
        ++global_state.buf_start;
        global_state.buf_start &= 0x1f;
    }
}

void global_state_puts(const char *msg)
{
    while(*msg)
        global_state_putchar(*msg++);
}

void global_state_drop_message()
{
    global_state.buf_start = 0;
    global_state.buf_end = 0;
}

void sync_global_state()
{
    uint8_t module_mask = 
        ((get_extension(0) == 0x04) ? 0x0f : 0x00) |
        ((get_extension(1) == 0x04) ? 0xf0 : 0x00);

    global_state.state_bits_bytes[0] =
        ((OPORT&(1<<OUBIT_A))>>OUBIT_A)|
        ((OPORT&(1<<OUBIT_B))>>OUBIT_B<<1)|
        ((OPORT&(1<<OUBIT_C))>>OUBIT_C<<2)|
        ((OPORT&(1<<OUBIT_D))>>OUBIT_D<<3);

    global_state.state_bits_bytes[1] =
        (confirmed_input_state & module_mask) |
        (~get_actual_input_state_byte() & ~module_mask);
}

void save_registers_to_eeprom()
{
    eeprom_write_block(&global_configuration.hardware_config,
                       (unsigned char *)0 + 16, 48);
}

void zero_counters(uint16_t *save)
{
    uint8_t i;
    for(i=0; i<UNC_INPUTS_COUNT; i++) {
        /* if the counter is incremented both by pressing and releasing the
           button, *AND* the "keep parity" flag is set, then the counter's
           parity should reflect the current state of the button
         */
        uint8_t should_be_one =
            (
                ((global_configuration.react_bytes[i][3] & 0x70) == 0x70) &&
                (confirmed_input_state & (1<<i))
            ) ? 1 : 0; 
#if 0
        /* this proved to be not a good idea: it is better to give
           the requester the opportunity to check (by the counters'
           value) whether the button is pressed now or not
         */
        if(save) {
            save[i] = global_state.counters[i] - should_be_one;
        }
#else
        if(save) {
            save[i] = global_state.counters[i];
        }
#endif
        global_state.counters[i] = should_be_one;
    }
}

static void perform_extended_action(uint8_t regnum)
{
    uint8_t reg[8];
    static const uint32_t mults[] = { 1, 100, 6000, 360000 };
    static uint8_t recursion_protector = 0;

    if(regnum >= 48) return;
    eeprom_read_block(reg, (uint8_t *)128 + regnum * 8, 8);
    if(reg[0] != 0) return;  /* not a valid reaction register */

    ++recursion_protector;
    /* first, perform the immediate action, if any */
    /* 0xff means noOp, and 0xf0 would cause a recursion,
       so it is restricted in case there are 5 levels already */
    if(reg[7] != 0xff && (reg[7] != 0xf0 || recursion_protector < 5)) {
        perform_action(reg[7], reg[6]);
    } 

    /* now set up the delayed action, if any */
    /* this must be done AFTER the immediate action, as 
       the immediate one could be "cancel all delayed" */
    if(reg[3] != 0xff) {
        uint8_t i;
        for(i=0; i<MAX_DELAYED_ACTIONS; i++)
            if(delayed_actions[i].grp == 0xff) {
                /* here we go! */
                delayed_actions[i].grp = reg[3];
                delayed_actions[i].act = reg[2];
                delayed_action_counters[i] = reg[5] * mults[reg[4] & 0x03];
                break;
            }
    }
    --recursion_protector;
}

void perform_action(uint8_t grp, uint8_t action)
{
    static volatile uint8_t* ports[3] = { &OPORT, &OPORT1, &OPORT2 };
    static const uint8_t bits[3][4] = {
        { 1<<OUBIT_A,  1<<OUBIT_B,  1<<OUBIT_C,  1<<OUBIT_D  },
        { 1<<OUBIT1_A, 1<<OUBIT1_B, 1<<OUBIT1_C, 1<<OUBIT1_D },
        { 1<<OUBIT2_A, 1<<OUBIT2_B, 1<<OUBIT2_C, 1<<OUBIT2_D }
    };
    uint8_t i;

#if 0
    print_str("X:");
    print_hex_byte(grp);
    print_hex_byte(action);
    print_char(':');
    print_int(global_state.timemark);
    print_char('\n');
#endif

    switch(grp) {
    case 0:
    case 1:
    case 2:
        for(i = 0; i < 4; i++) {
            switch((action >> (2 * i)) & 3 /* least 2 bits */) {
                case 0:   /* 00: no-op */
                    break;
                case 1:   /* 01: on */
                    *ports[grp] |= bits[grp][i];
                    break;
                case 2:   /* 10: off */
                    *ports[grp] &= ~bits[grp][i];
                    break;
                case 3:   /* 11: switch */
                    *ports[grp] ^= bits[grp][i];
                    break;
            }
        }
        break;
    case 0xf0:
        perform_extended_action(action);
        break;
    case 0xfd: /* output a message */
        print_str("A:");
        print_hex_byte(action);
        print_char('\n');
        break;
    case 0xfe: /* cancel all delays */
        cancel_delayed_actions(action);
        break;
    case 0xff: /* no operation */
    default:   /* actually an error */
        break;
    } 
}

static void process_input_change(uint8_t i,
                                 uint8_t react_idx,
                                 uint8_t cntr_mask,
                                 uint8_t out_mask,
                                 uint8_t out_char)
{
    static char in_msg[] = "L+0\n";
    perform_action(
        global_configuration.react_bytes[i][2], 
        global_configuration.react_bytes[i][react_idx]
    );
    if(global_configuration.react_bytes[i][3] & cntr_mask)
        global_state.counters[i]++;
    if(global_configuration.react_bytes[i][3] & out_mask) {
        in_msg[1] = out_char;
        in_msg[2] = '0' + i;
        print_str(in_msg);
    }
} 

void do_input_confirmation()
{
    uint8_t input_diff =
        get_actual_input_state_byte() ^ confirmed_input_state;
    uint8_t i;
    for(i = (get_extension(0) == 0x04) ? 0 : 4;
        i < UNC_INPUTS_COUNT; i++)
    {
        if(global_configuration.react_bytes[i][3] & 0x80) /* analog mode */
            continue;
        if(i>=4 && (get_extension(1) != 0x04))
            break;
        if(input_diff & (1<<i)) {
            if(GSTMRK - input_timestamps[i] > 
                global_configuration.user_config_bytes[0])
            {
                /* yes it has really changed */
                confirmed_input_state ^= (1<<i);
                /* now handle the I-register */
                if(confirmed_input_state & (1<<i)) {
                    /* pressed */
                    process_input_change(i, 0, 0x20, 0x04, '+');
                } else {
                    /* released */
                    process_input_change(i, 1, 0x40, 0x08, '-');
                }

                /* reset timestamp (yes, we're paranoid enough :) */
                input_timestamps[i] = GSTMRK;
            }
        } else { /* no difference */
            input_timestamps[i] = GSTMRK;
        }
    }
}

void process_delayed_actions()
{
    uint8_t i, grp;
    for(i=0; i<MAX_DELAYED_ACTIONS; i++)
        if((grp = delayed_actions[i].grp) != 0xff) {
            if(--(delayed_action_counters[i]) == 0) {
                delayed_actions[i].grp = 0xff;
                perform_action(grp, delayed_actions[i].act);
            }
        }
}


void init_timer()
{

    /* TCCR1B=0x05; // CLK/1024 */
    /* TCCR1B|=(1<<CS12)|(1<<CS10);  */
    /* TCCR1B|=(1<<CS12)|(1<<CS10);  */
    /* TCCR1B|=(1<<CS12); //CLK/256 */
    /* TCCR1B |= (1<<CS11); */ /* CLK/8 */


    /* CLK/8 */
    TCCR1B = 2;
    //OCR1AH=58;
    //OCR1AL=152;  /* 10 mS */
    //OCR1A = 15000;
    OCR1A = *((uint16_t*)&(global_configuration.hardware_config_bytes[2]));

#if 0
    TCCR1B = 4;  /* CLK/256 */
    OCR1AH=1;
    OCR1AL=213;  /* 10 mS */

    TCCR1B = 5;  /* CLK/1024 */
    OCR1AH=0;
    OCR1AL=117;  /* 10 mS */
#endif

    TIMSK|=(1<<OCIE1A); // Allow the interrupt upon match
}
